package scanner

import file.FilePosition
import org.junit.Test
import token.Token

import java.io.InputStream

import org.junit.Assert.assertTrue
import testutil.TestUtilities.createInputStreamFromString

class LispScannerLineColumnTest {

    private class LineColumn {

        private var line: Int = 0
        private var column: Int = 0

        fun isEqual(position: FilePosition): Boolean {
            return this.line == position.lineNumber && this.column == position.columnNumber
        }

        companion object {

            fun create(line: Int, column: Int): LineColumn {
                val lineColumn = LineColumn()
                lineColumn.line = line
                lineColumn.column = column

                return lineColumn
            }
        }
    }

    private fun assertTokenLineAndColumnsMatch(input: String, expectedLineColumnList: Array<LineColumn>) {
        val stringInputStream = createInputStreamFromString(input)
        val lispScanner = LispScanner(stringInputStream, "testFile")

        for (lineColumn in expectedLineColumnList) {
            val nextToken = lispScanner.nextToken
            assertTrue(lineColumn.isEqual(nextToken.position))
        }
    }

    @Test
    fun givenNothing_RecordsCorrectEofLocation() {
        val input = ""
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 0))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenSimpleString_RecordsCorrectLocation() {
        val input = "\"string\""
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenStringWithTrailingSpace_RecordsCorrectLocation() {
        val input = "\"string\" "
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenIdentifier_RecordsCorrectLocation() {
        val input = "identifier"
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenIdentifierWithTrailingSpace_RecordsCorrectLocation() {
        val input = "identifier "
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenNumber_RecordsCorrectLocation() {
        val input = "123456789"
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenNumberWithTrailingSpace_RecordsCorrectLocation() {
        val input = "123456789 "
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenMultipleStrings_RecordsCorrectLocations() {
        val input = "\"string1\" \n \"string2 \n with newline\" \n  \"string3\""
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1), LineColumn.create(2, 2), LineColumn.create(4, 3))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenQuotedList_RecordsCorrectLocations() {
        val input = "'(1 2 3 4 5)"
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1),
                                              LineColumn.create(1, 2),
                                              LineColumn.create(1, 3),
                                              LineColumn.create(1, 5),
                                              LineColumn.create(1, 7),
                                              LineColumn.create(1, 9),
                                              LineColumn.create(1, 11),
                                              LineColumn.create(1, 12))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenListSpanningMultipleLines_RecordsCorrectLocations() {
        val input = " ( 1 2 \n 3 4 \n5 ) "
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 2),
                                              LineColumn.create(1, 4),
                                              LineColumn.create(1, 6),
                                              LineColumn.create(2, 2),
                                              LineColumn.create(2, 4),
                                              LineColumn.create(3, 1),
                                              LineColumn.create(3, 3))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }

    @Test
    fun givenCommentImmediatelyFollowingNumber_RecordsCorrectLocations() {
        val input = "12;comment\n34"
        val expectedLinesAndColumns = arrayOf(LineColumn.create(1, 1), LineColumn.create(2, 1))

        assertTokenLineAndColumnsMatch(input, expectedLinesAndColumns)
    }
}
